import time

from sqlglot.optimizer import optimize

TPCH_SCHEMA = {
    "lineitem": {
        "l_orderkey": "bigint",
        "l_partkey": "bigint",
        "l_suppkey": "bigint",
        "l_linenumber": "bigint",
        "l_quantity": "double",
        "l_extendedprice": "double",
        "l_discount": "double",
        "l_tax": "double",
        "l_returnflag": "string",
        "l_linestatus": "string",
        "l_shipdate": "string",
        "l_commitdate": "string",
        "l_receiptdate": "string",
        "l_shipinstruct": "string",
        "l_shipmode": "string",
        "l_comment": "string",
    },
    "orders": {
        "o_orderkey": "bigint",
        "o_custkey": "bigint",
        "o_orderstatus": "string",
        "o_totalprice": "double",
        "o_orderdate": "string",
        "o_orderpriority": "string",
        "o_clerk": "string",
        "o_shippriority": "int",
        "o_comment": "string",
    },
    "customer": {
        "c_custkey": "bigint",
        "c_name": "string",
        "c_address": "string",
        "c_nationkey": "bigint",
        "c_phone": "string",
        "c_acctbal": "double",
        "c_mktsegment": "string",
        "c_comment": "string",
    },
    "part": {
        "p_partkey": "bigint",
        "p_name": "string",
        "p_mfgr": "string",
        "p_brand": "string",
        "p_type": "string",
        "p_size": "int",
        "p_container": "string",
        "p_retailprice": "double",
        "p_comment": "string",
    },
    "supplier": {
        "s_suppkey": "bigint",
        "s_name": "string",
        "s_address": "string",
        "s_nationkey": "bigint",
        "s_phone": "string",
        "s_acctbal": "double",
        "s_comment": "string",
    },
    "partsupp": {
        "ps_partkey": "bigint",
        "ps_suppkey": "bigint",
        "ps_availqty": "int",
        "ps_supplycost": "double",
        "ps_comment": "string",
    },
    "nation": {
        "n_nationkey": "bigint",
        "n_name": "string",
        "n_regionkey": "bigint",
        "n_comment": "string",
    },
    "region": {
        "r_regionkey": "bigint",
        "r_name": "string",
        "r_comment": "string",
    },
}

TPCDS_SCHEMA = {
    "catalog_sales": {
        "cs_sold_date_sk": "bigint",
        "cs_sold_time_sk": "bigint",
        "cs_ship_date_sk": "bigint",
        "cs_bill_customer_sk": "bigint",
        "cs_bill_cdemo_sk": "bigint",
        "cs_bill_hdemo_sk": "bigint",
        "cs_bill_addr_sk": "bigint",
        "cs_ship_customer_sk": "bigint",
        "cs_ship_cdemo_sk": "bigint",
        "cs_ship_hdemo_sk": "bigint",
        "cs_ship_addr_sk": "bigint",
        "cs_call_center_sk": "bigint",
        "cs_catalog_page_sk": "bigint",
        "cs_ship_mode_sk": "bigint",
        "cs_warehouse_sk": "bigint",
        "cs_item_sk": "bigint",
        "cs_promo_sk": "bigint",
        "cs_order_number": "bigint",
        "cs_quantity": "bigint",
        "cs_wholesale_cost": "double",
        "cs_list_price": "double",
        "cs_sales_price": "double",
        "cs_ext_discount_amt": "double",
        "cs_ext_sales_price": "double",
        "cs_ext_wholesale_cost": "double",
        "cs_ext_list_price": "double",
        "cs_ext_tax": "double",
        "cs_coupon_amt": "double",
        "cs_ext_ship_cost": "double",
        "cs_net_paid": "double",
        "cs_net_paid_inc_tax": "double",
        "cs_net_paid_inc_ship": "double",
        "cs_net_paid_inc_ship_tax": "double",
        "cs_net_profit": "double",
    },
    "catalog_returns": {
        "cr_returned_date_sk": "bigint",
        "cr_returned_time_sk": "bigint",
        "cr_item_sk": "bigint",
        "cr_refunded_customer_sk": "bigint",
        "cr_refunded_cdemo_sk": "bigint",
        "cr_refunded_hdemo_sk": "bigint",
        "cr_refunded_addr_sk": "bigint",
        "cr_returning_customer_sk": "bigint",
        "cr_returning_cdemo_sk": "bigint",
        "cr_returning_hdemo_sk": "bigint",
        "cr_returning_addr_sk": "bigint",
        "cr_call_center_sk": "bigint",
        "cr_catalog_page_sk": "bigint",
        "cr_ship_mode_sk": "bigint",
        "cr_warehouse_sk": "bigint",
        "cr_reason_sk": "bigint",
        "cr_order_number": "bigint",
        "cr_return_quantity": "bigint",
        "cr_return_amount": "double",
        "cr_return_tax": "double",
        "cr_return_amt_inc_tax": "double",
        "cr_fee": "double",
        "cr_return_ship_cost": "double",
        "cr_refunded_cash": "double",
        "cr_reversed_charge": "double",
        "cr_store_credit": "double",
        "cr_net_loss": "double",
    },
    "inventory": {
        "inv_date_sk": "bigint",
        "inv_item_sk": "bigint",
        "inv_warehouse_sk": "bigint",
        "inv_quantity_on_hand": "bigint",
    },
    "store_sales": {
        "ss_sold_date_sk": "bigint",
        "ss_sold_time_sk": "bigint",
        "ss_item_sk": "bigint",
        "ss_customer_sk": "bigint",
        "ss_cdemo_sk": "bigint",
        "ss_hdemo_sk": "bigint",
        "ss_addr_sk": "bigint",
        "ss_store_sk": "bigint",
        "ss_promo_sk": "bigint",
        "ss_ticket_number": "bigint",
        "ss_quantity": "bigint",
        "ss_wholesale_cost": "double",
        "ss_list_price": "double",
        "ss_sales_price": "double",
        "ss_ext_discount_amt": "double",
        "ss_ext_sales_price": "double",
        "ss_ext_wholesale_cost": "double",
        "ss_ext_list_price": "double",
        "ss_ext_tax": "double",
        "ss_coupon_amt": "double",
        "ss_net_paid": "double",
        "ss_net_paid_inc_tax": "double",
        "ss_net_profit": "double",
    },
    "store_returns": {
        "sr_returned_date_sk": "bigint",
        "sr_return_time_sk": "bigint",
        "sr_item_sk": "bigint",
        "sr_customer_sk": "bigint",
        "sr_cdemo_sk": "bigint",
        "sr_hdemo_sk": "bigint",
        "sr_addr_sk": "bigint",
        "sr_store_sk": "bigint",
        "sr_reason_sk": "bigint",
        "sr_ticket_number": "bigint",
        "sr_return_quantity": "bigint",
        "sr_return_amt": "double",
        "sr_return_tax": "double",
        "sr_return_amt_inc_tax": "double",
        "sr_fee": "double",
        "sr_return_ship_cost": "double",
        "sr_refunded_cash": "double",
        "sr_reversed_charge": "double",
        "sr_store_credit": "double",
        "sr_net_loss": "double",
    },
    "web_sales": {
        "ws_sold_date_sk": "bigint",
        "ws_sold_time_sk": "bigint",
        "ws_ship_date_sk": "bigint",
        "ws_item_sk": "bigint",
        "ws_bill_customer_sk": "bigint",
        "ws_bill_cdemo_sk": "bigint",
        "ws_bill_hdemo_sk": "bigint",
        "ws_bill_addr_sk": "bigint",
        "ws_ship_customer_sk": "bigint",
        "ws_ship_cdemo_sk": "bigint",
        "ws_ship_hdemo_sk": "bigint",
        "ws_ship_addr_sk": "bigint",
        "ws_web_page_sk": "bigint",
        "ws_web_site_sk": "bigint",
        "ws_ship_mode_sk": "bigint",
        "ws_warehouse_sk": "bigint",
        "ws_promo_sk": "bigint",
        "ws_order_number": "bigint",
        "ws_quantity": "bigint",
        "ws_wholesale_cost": "double",
        "ws_list_price": "double",
        "ws_sales_price": "double",
        "ws_ext_discount_amt": "double",
        "ws_ext_sales_price": "double",
        "ws_ext_wholesale_cost": "double",
        "ws_ext_list_price": "double",
        "ws_ext_tax": "double",
        "ws_coupon_amt": "double",
        "ws_ext_ship_cost": "double",
        "ws_net_paid": "double",
        "ws_net_paid_inc_tax": "double",
        "ws_net_paid_inc_ship": "double",
        "ws_net_paid_inc_ship_tax": "double",
        "ws_net_profit": "double",
    },
    "web_returns": {
        "wr_returned_date_sk": "bigint",
        "wr_returned_time_sk": "bigint",
        "wr_item_sk": "bigint",
        "wr_refunded_customer_sk": "bigint",
        "wr_refunded_cdemo_sk": "bigint",
        "wr_refunded_hdemo_sk": "bigint",
        "wr_refunded_addr_sk": "bigint",
        "wr_returning_customer_sk": "bigint",
        "wr_returning_cdemo_sk": "bigint",
        "wr_returning_hdemo_sk": "bigint",
        "wr_returning_addr_sk": "bigint",
        "wr_web_page_sk": "bigint",
        "wr_reason_sk": "bigint",
        "wr_order_number": "bigint",
        "wr_return_quantity": "bigint",
        "wr_return_amt": "double",
        "wr_return_tax": "double",
        "wr_return_amt_inc_tax": "double",
        "wr_fee": "double",
        "wr_return_ship_cost": "double",
        "wr_refunded_cash": "double",
        "wr_reversed_charge": "double",
        "wr_account_credit": "double",
        "wr_net_loss": "double",
    },
    "call_center": {
        "cc_call_center_sk": "bigint",
        "cc_call_center_id": "string",
        "cc_rec_start_date": "string",
        "cc_rec_end_date": "string",
        "cc_closed_date_sk": "bigint",
        "cc_open_date_sk": "bigint",
        "cc_name": "string",
        "cc_class": "string",
        "cc_employees": "bigint",
        "cc_sq_ft": "bigint",
        "cc_hours": "string",
        "cc_manager": "string",
        "cc_mkt_id": "bigint",
        "cc_mkt_class": "string",
        "cc_mkt_desc": "string",
        "cc_market_manager": "string",
        "cc_division": "bigint",
        "cc_division_name": "string",
        "cc_company": "bigint",
        "cc_company_name": "string",
        "cc_street_number": "string",
        "cc_street_name": "string",
        "cc_street_type": "string",
        "cc_suite_number": "string",
        "cc_city": "string",
        "cc_county": "string",
        "cc_state": "string",
        "cc_zip": "string",
        "cc_country": "string",
        "cc_gmt_offset": "double",
        "cc_tax_percentage": "double",
    },
    "catalog_page": {
        "cp_catalog_page_sk": "bigint",
        "cp_catalog_page_id": "string",
        "cp_start_date_sk": "bigint",
        "cp_end_date_sk": "bigint",
        "cp_department": "string",
        "cp_catalog_number": "bigint",
        "cp_catalog_page_number": "bigint",
        "cp_description": "string",
        "cp_type": "string",
    },
    "customer": {
        "c_customer_sk": "bigint",
        "c_customer_id": "string",
        "c_current_cdemo_sk": "bigint",
        "c_current_hdemo_sk": "bigint",
        "c_current_addr_sk": "bigint",
        "c_first_shipto_date_sk": "bigint",
        "c_first_sales_date_sk": "bigint",
        "c_salutation": "string",
        "c_first_name": "string",
        "c_last_name": "string",
        "c_preferred_cust_flag": "string",
        "c_birth_day": "bigint",
        "c_birth_month": "bigint",
        "c_birth_year": "bigint",
        "c_birth_country": "string",
        "c_login": "string",
        "c_email_address": "string",
        "c_last_review_date": "string",
    },
    "customer_address": {
        "ca_address_sk": "bigint",
        "ca_address_id": "string",
        "ca_street_number": "string",
        "ca_street_name": "string",
        "ca_street_type": "string",
        "ca_suite_number": "string",
        "ca_city": "string",
        "ca_county": "string",
        "ca_state": "string",
        "ca_zip": "string",
        "ca_country": "string",
        "ca_gmt_offset": "double",
        "ca_location_type": "string",
    },
    "customer_demographics": {
        "cd_demo_sk": "bigint",
        "cd_gender": "string",
        "cd_marital_status": "string",
        "cd_education_status": "string",
        "cd_purchase_estimate": "bigint",
        "cd_credit_rating": "string",
        "cd_dep_count": "bigint",
        "cd_dep_employed_count": "bigint",
        "cd_dep_college_count": "bigint",
    },
    "date_dim": {
        "d_date_sk": "bigint",
        "d_date_id": "string",
        "d_date": "string",
        "d_month_seq": "bigint",
        "d_week_seq": "bigint",
        "d_quarter_seq": "bigint",
        "d_year": "bigint",
        "d_dow": "bigint",
        "d_moy": "bigint",
        "d_dom": "bigint",
        "d_qoy": "bigint",
        "d_fy_year": "bigint",
        "d_fy_quarter_seq": "bigint",
        "d_fy_week_seq": "bigint",
        "d_day_name": "string",
        "d_quarter_name": "string",
        "d_holiday": "string",
        "d_weekend": "string",
        "d_following_holiday": "string",
        "d_first_dom": "bigint",
        "d_last_dom": "bigint",
        "d_same_day_ly": "bigint",
        "d_same_day_lq": "bigint",
        "d_current_day": "string",
        "d_current_week": "string",
        "d_current_month": "string",
        "d_current_quarter": "string",
        "d_current_year": "string",
    },
    "household_demographics": {
        "hd_demo_sk": "bigint",
        "hd_income_band_sk": "bigint",
        "hd_buy_potential": "string",
        "hd_dep_count": "bigint",
        "hd_vehicle_count": "bigint",
    },
    "income_band": {
        "ib_income_band_sk": "bigint",
        "ib_lower_bound": "bigint",
        "ib_upper_bound": "bigint",
    },
    "item": {
        "i_item_sk": "bigint",
        "i_item_id": "string",
        "i_rec_start_date": "string",
        "i_rec_end_date": "string",
        "i_item_desc": "string",
        "i_current_price": "double",
        "i_wholesale_cost": "double",
        "i_brand_id": "bigint",
        "i_brand": "string",
        "i_class_id": "bigint",
        "i_class": "string",
        "i_category_id": "bigint",
        "i_category": "string",
        "i_manufact_id": "bigint",
        "i_manufact": "string",
        "i_size": "string",
        "i_formulation": "string",
        "i_color": "string",
        "i_units": "string",
        "i_container": "string",
        "i_manager_id": "bigint",
        "i_product_name": "string",
    },
    "promotion": {
        "p_promo_sk": "bigint",
        "p_promo_id": "string",
        "p_start_date_sk": "bigint",
        "p_end_date_sk": "bigint",
        "p_item_sk": "bigint",
        "p_cost": "double",
        "p_response_target": "bigint",
        "p_promo_name": "string",
        "p_channel_dmail": "string",
        "p_channel_email": "string",
        "p_channel_catalog": "string",
        "p_channel_tv": "string",
        "p_channel_radio": "string",
        "p_channel_press": "string",
        "p_channel_event": "string",
        "p_channel_demo": "string",
        "p_channel_details": "string",
        "p_purpose": "string",
        "p_discount_active": "string",
    },
    "reason": {"r_reason_sk": "bigint", "r_reason_id": "string", "r_reason_desc": "string"},
    "ship_mode": {
        "sm_ship_mode_sk": "bigint",
        "sm_ship_mode_id": "string",
        "sm_type": "string",
        "sm_code": "string",
        "sm_carrier": "string",
        "sm_contract": "string",
    },
    "store": {
        "s_store_sk": "bigint",
        "s_store_id": "string",
        "s_rec_start_date": "string",
        "s_rec_end_date": "string",
        "s_closed_date_sk": "bigint",
        "s_store_name": "string",
        "s_number_employees": "bigint",
        "s_floor_space": "bigint",
        "s_hours": "string",
        "s_manager": "string",
        "s_market_id": "bigint",
        "s_geography_class": "string",
        "s_market_desc": "string",
        "s_market_manager": "string",
        "s_division_id": "bigint",
        "s_division_name": "string",
        "s_company_id": "bigint",
        "s_company_name": "string",
        "s_street_number": "string",
        "s_street_name": "string",
        "s_street_type": "string",
        "s_suite_number": "string",
        "s_city": "string",
        "s_county": "string",
        "s_state": "string",
        "s_zip": "string",
        "s_country": "string",
        "s_gmt_offset": "double",
        "s_tax_precentage": "double",
    },
    "time_dim": {
        "t_time_sk": "bigint",
        "t_time_id": "string",
        "t_time": "bigint",
        "t_hour": "bigint",
        "t_minute": "bigint",
        "t_second": "bigint",
        "t_am_pm": "string",
        "t_shift": "string",
        "t_sub_shift": "string",
        "t_meal_time": "string",
    },
    "warehouse": {
        "w_warehouse_sk": "bigint",
        "w_warehouse_id": "string",
        "w_warehouse_name": "string",
        "w_warehouse_sq_ft": "bigint",
        "w_street_number": "string",
        "w_street_name": "string",
        "w_street_type": "string",
        "w_suite_number": "string",
        "w_city": "string",
        "w_county": "string",
        "w_state": "string",
        "w_zip": "string",
        "w_country": "string",
        "w_gmt_offset": "double",
    },
    "web_page": {
        "wp_web_page_sk": "bigint",
        "wp_web_page_id": "string",
        "wp_rec_start_date": "string",
        "wp_rec_end_date": "string",
        "wp_creation_date_sk": "bigint",
        "wp_access_date_sk": "bigint",
        "wp_autogen_flag": "string",
        "wp_customer_sk": "bigint",
        "wp_url": "string",
        "wp_type": "string",
        "wp_char_count": "bigint",
        "wp_link_count": "bigint",
        "wp_image_count": "bigint",
        "wp_max_ad_count": "bigint",
    },
    "web_site": {
        "web_site_sk": "bigint",
        "web_site_id": "string",
        "web_rec_start_date": "string",
        "web_rec_end_date": "string",
        "web_name": "string",
        "web_open_date_sk": "bigint",
        "web_close_date_sk": "bigint",
        "web_class": "string",
        "web_manager": "string",
        "web_mkt_id": "bigint",
        "web_mkt_class": "string",
        "web_mkt_desc": "string",
        "web_market_manager": "string",
        "web_company_id": "bigint",
        "web_company_name": "string",
        "web_street_number": "string",
        "web_street_name": "string",
        "web_street_type": "string",
        "web_suite_number": "string",
        "web_city": "string",
        "web_county": "string",
        "web_state": "string",
        "web_zip": "string",
        "web_country": "string",
        "web_gmt_offset": "string",
        "web_tax_percentage": "double",
    },
}


def rewrite_fixtures(in_path, out_path, schema, num, kind):
    with open(out_path, "w", encoding="utf-8") as fixture:
        for i in range(num):
            i = i + 1
            with open(in_path.format(i=i), encoding="utf-8") as file:
                original = "\n".join(
                    line.rstrip()
                    for line in file.read().split(";")[0].split("\n")
                    if not line.startswith("--")
                )
                original = original.replace("`", '"').strip()
                now = time.time()
                try:
                    optimized = optimize(original, schema=schema)
                except Exception as e:
                    print("****", i, e, "****")
                    continue

                fixture.write(
                    f"""--------------------------------------
-- TPC-{kind} {i}
--------------------------------------
{original};
{optimized.sql(pretty=True)};

"""
                )
                print(i, time.time() - now)


rewrite_fixtures(
    "/home/toby/dev/tpch/{i}.sql",
    "/home/toby/dev/sqlglot/tests/fixtures/optimizer/tpc-h/tpc-h.sql",
    TPCH_SCHEMA,
    22,
    "H",
)

rewrite_fixtures(
    "/home/toby/dev/tpcds/query{i}.sql",
    "/home/toby/dev/sqlglot/tests/fixtures/optimizer/tpc-ds/tpc-ds.sql",
    TPCDS_SCHEMA,
    99,
    "DS",
)



from __future__ import annotations

import typing as t

from sqlglot import expressions as exp
from sqlglot import parser
from sqlglot.dialects.dialect import (
    build_date_delta,
    build_formatted_time,
    build_timetostr_or_tochar,
)
from sqlglot.helper import seq_get
from sqlglot.tokens import TokenType

if t.TYPE_CHECKING:
    from sqlglot.dialects.dialect import DialectType

DATE_DELTA = t.Union[exp.DateAdd, exp.DateSub]


def to_char_is_numeric_handler(args: list, dialect: DialectType) -> exp.TimeToStr | exp.ToChar:
    expression = build_timetostr_or_tochar(args, dialect)
    fmt = seq_get(args, 1)

    if fmt and isinstance(expression, exp.ToChar) and fmt.is_string and "#" in fmt.name:
        # Only mark as numeric if format is a literal containing #
        expression.set("is_numeric", True)

    return expression


def build_date_delta_with_cast_interval(
    expression_class: type[DATE_DELTA],
) -> t.Callable[[list[exp.Expr]], exp.Expr]:
    fallback_builder = build_date_delta(expression_class)

    def _builder(args):
        if len(args) == 2:
            date_arg, interval_arg = args

            if (
                isinstance(interval_arg, exp.Cast)
                and isinstance(interval_arg.to, exp.DataType)
                and isinstance(interval_arg.to.this, exp.Interval)
            ):
                return expression_class(
                    this=date_arg,
                    expression=interval_arg.this,
                    unit=interval_arg.to.this.unit,
                )

            return expression_class(this=date_arg, expression=interval_arg)

        return fallback_builder(args)

    return _builder


def datetype_handler(args: list[exp.Expr], dialect: DialectType) -> exp.Expr:
    from sqlglot.dialects.dialect import Dialect

    year, month, day = args

    if all(isinstance(arg, exp.Literal) and arg.is_int for arg in (year, month, day)):
        date_str = f"{int(year.this):04d}-{int(month.this):02d}-{int(day.this):02d}"
        return exp.Date(this=exp.Literal.string(date_str))

    dialect = Dialect.get_or_raise(dialect)

    return exp.Cast(
        this=exp.Concat(
            expressions=[
                year,
                exp.Literal.string("-"),
                month,
                exp.Literal.string("-"),
                day,
            ],
            coalesce=dialect.CONCAT_COALESCE,
        ),
        to=exp.DType.DATE.into_expr(),
    )


class DremioParser(parser.Parser):
    LOG_DEFAULTS_TO_LN = True

    TABLE_ALIAS_TOKENS = parser.Parser.TABLE_ALIAS_TOKENS | {
        TokenType.ANTI,
        TokenType.SEMI,
    }

    NO_PAREN_FUNCTION_PARSERS = {
        **parser.Parser.NO_PAREN_FUNCTION_PARSERS,
        "CURRENT_DATE_UTC": lambda self: self._parse_current_date_utc(),
    }

    FUNCTIONS = {
        **parser.Parser.FUNCTIONS,
        "ARRAY_GENERATE_RANGE": exp.GenerateSeries.from_arg_list,
        "BIT_AND": exp.BitwiseAndAgg.from_arg_list,
        "BIT_OR": exp.BitwiseOrAgg.from_arg_list,
        "DATE_ADD": build_date_delta_with_cast_interval(exp.DateAdd),
        "DATE_FORMAT": build_formatted_time(exp.TimeToStr),
        "DATE_SUB": build_date_delta_with_cast_interval(exp.DateSub),
        "REGEXP_MATCHES": exp.RegexpLike.from_arg_list,
        "REPEATSTR": exp.Repeat.from_arg_list,
        "TO_CHAR": to_char_is_numeric_handler,
        "TO_DATE": build_formatted_time(exp.TsOrDsToDate),
        "DATE_PART": exp.Extract.from_arg_list,
        "DATETYPE": datetype_handler,
    }

    VERSION_PHRASES = {
        **parser.Parser.VERSION_PHRASES,
        ("AT", "TIMESTAMP"): "TIMESTAMP",
        ("AT", "SNAPSHOT"): "VERSION",
    }

    def _parse_current_date_utc(self) -> exp.Cast:
        if self._match(TokenType.L_PAREN):
            self._match_r_paren()

        return exp.Cast(
            this=exp.AtTimeZone(
                this=exp.CurrentTimestamp(),
                zone=exp.Literal.string("UTC"),
            ),
            to=exp.DType.DATE.into_expr(),
        )



from __future__ import annotations

from sqlglot.typing.redshift import EXPRESSION_METADATA
from sqlglot.dialects.dialect import NormalizationStrategy
from sqlglot.dialects.postgres import Postgres
from sqlglot.generators.redshift import RedshiftGenerator
from sqlglot.parsers.redshift import RedshiftParser
from sqlglot.tokens import TokenType


class Redshift(Postgres):
    # https://docs.aws.amazon.com/redshift/latest/dg/r_names.html
    NORMALIZATION_STRATEGY = NormalizationStrategy.CASE_INSENSITIVE

    NORMALIZE_NOT_NULL = True

    EXPRESSION_METADATA = EXPRESSION_METADATA.copy()
    SUPPORTS_USER_DEFINED_TYPES = False
    INDEX_OFFSET = 0
    COPY_PARAMS_ARE_CSV = False
    HEX_LOWERCASE = True
    HAS_DISTINCT_ARRAY_CONSTRUCTORS = True
    COALESCE_COMPARISON_NON_STANDARD = True
    REGEXP_EXTRACT_POSITION_OVERFLOW_RETURNS_NULL = False
    ARRAY_FUNCS_PROPAGATES_NULLS = True

    # ref: https://docs.aws.amazon.com/redshift/latest/dg/r_FORMAT_strings.html
    TIME_FORMAT = "'YYYY-MM-DD HH24:MI:SS'"

    TIME_MAPPING = {
        **Postgres.TIME_MAPPING,
        "MON": "%b",
        "MONTH": "%B",
    }

    Parser = RedshiftParser

    class Tokenizer(Postgres.Tokenizer):
        BIT_STRINGS = []
        HEX_STRINGS = []
        STRING_ESCAPES = ["\\", "'"]

        KEYWORDS = {
            **Postgres.Tokenizer.KEYWORDS,
            "(+)": TokenType.JOIN_MARKER,
            "BINARY VARYING": TokenType.VARBINARY,
            "CURRENT_USER_ID": TokenType.CURRENT_USER_ID,
            "HLLSKETCH": TokenType.HLLSKETCH,
            "MINUS": TokenType.EXCEPT,
            "SUPER": TokenType.SUPER,
            "TOP": TokenType.TOP,
            "UNLOAD": TokenType.COMMAND,
            "USER": TokenType.CURRENT_USER,
            "VARBYTE": TokenType.VARBINARY,
        }
        KEYWORDS.pop("VALUES")

        # Redshift allows # to appear as a table identifier prefix
        SINGLE_TOKENS = Postgres.Tokenizer.SINGLE_TOKENS.copy()
        SINGLE_TOKENS.pop("#")

    Generator = RedshiftGenerator



#!/usr/bin/env python3
"""Compare two benchmark JSON files and output a markdown table with diff indicators."""

import json
import sys


def _fmt_time(seconds):
    if seconds >= 1:
        return f"{seconds:.2f}s"
    if seconds >= 1e-3:
        return f"{seconds * 1e3:.1f}ms"
    return f"{seconds * 1e6:.0f}us"


def _indicator(ratio):
    """Return an emoji indicator based on the speedup/slowdown ratio (pr_time / main_time)."""
    if ratio <= 0.95:
        return "\U0001f7e2\U0001f7e2"  # 5%+ faster
    if ratio <= 0.97:
        return "\U0001f7e2"  # 3-5% faster
    if ratio <= 0.99:
        return "\U0001f7e9"  # 1-3% faster
    if ratio <= 1.01:
        return "\u26aa"  # no significant change
    if ratio <= 1.03:
        return "\U0001f7e7"  # 1-3% slower
    if ratio <= 1.05:
        return "\U0001f534"  # 3-5% slower
    return "\U0001f534\U0001f534"  # 5%+ slower


def _diff_text(ratio):
    pct = (ratio - 1.0) * 100
    if abs(pct) < 0.05:
        return "0.0%"
    if pct < 0:
        return f"{-pct:.1f}% faster"
    return f"{pct:.1f}% slower"


def compare(main_file, pr_file):
    with open(main_file) as f:
        main = json.load(f)
    with open(pr_file) as f:
        pr = json.load(f)

    # Collect all query names across both parsers
    parsers = set()
    queries = []
    seen_queries = set()
    for key in list(main.keys()) + list(pr.keys()):
        parser, query = key.split(":", 1)
        parsers.add(parser)
        if query not in seen_queries:
            queries.append(query)
            seen_queries.add(query)

    # Sort parsers: sqlglot first, then sqlglotc
    parser_order = sorted(parsers, key=lambda p: (p != "sqlglot", p != "sqlglotc", p))

    # Build table
    lines = []
    lines.append("## Benchmark Results\n")
    lines.append(
        "**Legend:** \U0001f7e2\U0001f7e2 = 5%+ faster | \U0001f7e2 = 3-5% faster | \U0001f7e9 = 1-3% faster | \u26aa = unchanged | \U0001f7e7 = 1-3% slower | \U0001f534 = 3-5% slower | \U0001f534\U0001f534 = 5%+ slower\n"
    )

    for parser in parser_order:
        display = (
            "sqlglot" if parser == "sqlglot" else "sqlglot[c]" if parser == "sqlglotc" else parser
        )
        lines.append(f"\n### {display}\n")
        lines.append("| Query | main | PR | diff |  |")
        lines.append("| ----- | ---: | ---: | ---: | --- |")

        for query in queries:
            key = f"{parser}:{query}"
            main_time = main.get(key)
            pr_time = pr.get(key)

            if main_time is None and pr_time is None:
                continue

            main_str = _fmt_time(main_time) if main_time else "N/A"
            pr_str = _fmt_time(pr_time) if pr_time else "N/A"

            if main_time and pr_time:
                ratio = pr_time / main_time
                diff_str = _diff_text(ratio)
                ind = _indicator(ratio)
            else:
                diff_str = "N/A"
                ind = ""

            lines.append(f"| {query} | {main_str} | {pr_str} | {diff_str} | {ind} |")

    lines.append("\n---\n*Comment `/benchmark` to re-run.*")

    return "\n".join(lines)


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: compare_benchmarks.py <main.json> <pr.json>", file=sys.stderr)
        sys.exit(1)
    print(compare(sys.argv[1], sys.argv[2]))



"""sqlglot expressions - math, trigonometry, and bitwise functions."""

from __future__ import annotations

from sqlglot.expressions.core import Expression, Func, AggFunc


# Trigonometric


class Acos(Expression, Func):
    pass


class Acosh(Expression, Func):
    pass


class Asin(Expression, Func):
    pass


class Asinh(Expression, Func):
    pass


class Atan(Expression, Func):
    arg_types = {"this": True, "expression": False}


class Atanh(Expression, Func):
    pass


class Atan2(Expression, Func):
    arg_types = {"this": True, "expression": True}


class Cos(Expression, Func):
    pass


class Cosh(Expression, Func):
    pass


class Cot(Expression, Func):
    pass


class Coth(Expression, Func):
    pass


class Csc(Expression, Func):
    pass


class Csch(Expression, Func):
    pass


class Degrees(Expression, Func):
    pass


class Radians(Expression, Func):
    pass


class Sec(Expression, Func):
    pass


class Sech(Expression, Func):
    pass


class Sin(Expression, Func):
    pass


class Sinh(Expression, Func):
    pass


class Tan(Expression, Func):
    pass


class Tanh(Expression, Func):
    pass


# Geometric distance / similarity


class CosineDistance(Expression, Func):
    arg_types = {"this": True, "expression": True}


class DotProduct(Expression, Func):
    arg_types = {"this": True, "expression": True}


class EuclideanDistance(Expression, Func):
    arg_types = {"this": True, "expression": True}


class JarowinklerSimilarity(Expression, Func):
    arg_types = {
        "this": True,
        "expression": True,
        "case_insensitive": False,
        "integer_scale": False,
    }


class ManhattanDistance(Expression, Func):
    arg_types = {"this": True, "expression": True}


# Basic arithmetic / math


class Abs(Expression, Func):
    pass


class Cbrt(Expression, Func):
    pass


class Ceil(Expression, Func):
    arg_types = {"this": True, "decimals": False, "to": False}
    _sql_names = ["CEIL", "CEILING"]


class Exp(Expression, Func):
    pass


class Factorial(Expression, Func):
    pass


class Floor(Expression, Func):
    arg_types = {"this": True, "decimals": False, "to": False}


class IsInf(Expression, Func):
    _sql_names = ["IS_INF", "ISINF"]


class IsNan(Expression, Func):
    _sql_names = ["IS_NAN", "ISNAN"]


class Ln(Expression, Func):
    pass


class Log(Expression, Func):
    arg_types = {"this": True, "expression": False}


class Negative(Expression, Func):
    pass


class Nanvl(Expression, Func):
    arg_types = {"this": True, "expression": True}


class Pi(Expression, Func):
    arg_types = {}


class Rint(Expression, Func):
    pass


class Round(Expression, Func):
    arg_types = {
        "this": True,
        "decimals": False,
        "truncate": False,
        "casts_non_integer_decimals": False,
    }


class Sign(Expression, Func):
    _sql_names = ["SIGN", "SIGNUM"]


class Sqrt(Expression, Func):
    pass


class Trunc(Expression, Func):
    arg_types = {"this": True, "decimals": False, "fractions_supported": False}
    _sql_names = ["TRUNC", "TRUNCATE"]


# Safe arithmetic


class SafeAdd(Expression, Func):
    arg_types = {"this": True, "expression": True}


class SafeDivide(Expression, Func):
    arg_types = {"this": True, "expression": True}


class SafeMultiply(Expression, Func):
    arg_types = {"this": True, "expression": True}


class SafeNegate(Expression, Func):
    pass


class SafeSubtract(Expression, Func):
    arg_types = {"this": True, "expression": True}


# Bitwise


class BitwiseAndAgg(Expression, AggFunc):
    pass


class BitwiseCount(Expression, Func):
    pass


class BitwiseOrAgg(Expression, AggFunc):
    pass


class BitwiseXorAgg(Expression, AggFunc):
    pass


class BitmapBitPosition(Expression, Func):
    pass


class BitmapBucketNumber(Expression, Func):
    pass


class BitmapConstructAgg(Expression, AggFunc):
    pass


class BitmapCount(Expression, Func):
    pass


class BitmapOrAgg(Expression, AggFunc):
    pass


class Booland(Expression, Func):
    arg_types = {"this": True, "expression": True, "round_input": False}


class Boolnot(Expression, Func):
    arg_types = {"this": True, "round_input": False}


class Boolor(Expression, Func):
    arg_types = {"this": True, "expression": True, "round_input": False}


class BoolxorAgg(Expression, AggFunc):
    pass


class Getbit(Expression, Func):
    _sql_names = ["GETBIT", "GET_BIT"]
    # zero_is_msb means the most significant bit is indexed 0
    arg_types = {"this": True, "expression": True, "zero_is_msb": False}



"""sqlglot expressions - array, map, struct, and table-valued functions."""

from __future__ import annotations


from sqlglot.expressions.core import (
    Expression,
    Expr,
    Func,
    Binary,
    Predicate,
    to_identifier,
)
from sqlglot.helper import trait
from sqlglot.expressions.query import UDTF


# Array creation / construction


class Array(Expression, Func):
    arg_types = {
        "expressions": False,
        "bracket_notation": False,
        "struct_name_inheritance": False,
    }
    is_var_len_args = True


class ArrayConstructCompact(Expression, Func):
    arg_types = {"expressions": False}
    is_var_len_args = True


class List(Expression, Func):
    arg_types = {"expressions": False}
    is_var_len_args = True


class ToArray(Expression, Func):
    pass


# Array manipulation


class ArrayAppend(Expression, Func):
    arg_types = {"this": True, "expression": True, "null_propagation": False}


class ArrayCompact(Expression, Func):
    pass


class ArrayConcat(Expression, Func):
    _sql_names = ["ARRAY_CONCAT", "ARRAY_CAT"]
    arg_types = {"this": True, "expressions": False, "null_propagation": False}
    is_var_len_args = True


class ArrayFilter(Expression, Func):
    arg_types = {"this": True, "expression": True}
    _sql_names = ["FILTER", "ARRAY_FILTER"]


class ArrayInsert(Expression, Func):
    arg_types = {"this": True, "position": True, "expression": True, "offset": False}


class ArrayPrepend(Expression, Func):
    arg_types = {"this": True, "expression": True, "null_propagation": False}


class ArrayRemove(Expression, Func):
    arg_types = {"this": True, "expression": True, "null_propagation": False}


class ArrayRemoveAt(Expression, Func):
    arg_types = {"this": True, "position": True}


class ArrayReverse(Expression, Func):
    pass


class ArraySlice(Expression, Func):
    arg_types = {"this": True, "start": True, "end": False, "step": False, "zero_based": False}


class ArraySort(Expression, Func):
    arg_types = {"this": True, "expression": False}


class Shuffle(Expression, Func):
    pass


class SortArray(Expression, Func):
    arg_types = {"this": True, "asc": False, "nulls_first": False}


# Array predicates / search


class ArrayAll(Expression, Func):
    arg_types = {"this": True, "expression": True}


class ArrayAny(Expression, Func):
    arg_types = {"this": True, "expression": True}


class ArrayContains(Expression, Binary, Predicate, Func):
    arg_types = {"this": True, "expression": True, "ensure_variant": False, "check_null": False}
    _sql_names = ["ARRAY_CONTAINS", "ARRAY_HAS"]


class ArrayContainsAll(Expression, Binary, Predicate, Func):
    _sql_names = ["ARRAY_CONTAINS_ALL", "ARRAY_HAS_ALL"]


class ArrayContainedBy(Expression, Binary, Predicate, Func):
    pass


class ArrayExcept(Expression, Func):
    arg_types = {"this": True, "expression": True, "is_multiset": False}


class ArrayIntersect(Expression, Func):
    arg_types = {"expressions": True, "is_multiset": False}
    is_var_len_args = True
    _sql_names = ["ARRAY_INTERSECT", "ARRAY_INTERSECTION"]


class ArrayOverlaps(Expression, Binary, Predicate, Func):
    arg_types = {"this": True, "expression": True, "null_safe": False}


class ArrayPosition(Expression, Binary, Func):
    arg_types = {"this": True, "expression": True, "zero_based": False}


# Array properties


class ArrayDistinct(Expression, Func):
    arg_types = {"this": True, "check_null": False}


class ArrayFirst(Expression, Func):
    arg_types = {"this": True, "expression": False}


class ArrayLast(Expression, Func):
    pass


class ArrayMax(Expression, Func):
    pass


class ArrayMin(Expression, Func):
    pass


class ArraySize(Expression, Func):
    arg_types = {"this": True, "expression": False}
    _sql_names = ["ARRAY_SIZE", "ARRAY_LENGTH"]


class ArraySum(Expression, Func):
    arg_types = {"this": True, "expression": False}


# Array conversion / utility


class ArraysZip(Expression, Func):
    arg_types = {"expressions": False}
    is_var_len_args = True


class ArrayToString(Expression, Func):
    arg_types = {
        "this": True,
        "expression": True,
        "null": False,
        "null_is_empty": False,
        "null_delim_is_null": False,
    }
    _sql_names = ["ARRAY_TO_STRING", "ARRAY_JOIN"]


class Flatten(Expression, Func):
    pass


class StringToArray(Expression, Func):
    arg_types = {"this": True, "expression": False, "null": False}
    _sql_names = ["STRING_TO_ARRAY", "SPLIT_BY_STRING"]


class StrtokToArray(Expression, Func):
    arg_types = {"this": True, "expression": False}


# Higher-order / lambda


class Apply(Expression, Func):
    arg_types = {"this": True, "expression": True}


class Reduce(Expression, Func):
    arg_types = {"this": True, "initial": True, "merge": True, "finish": False}


class Transform(Expression, Func):
    arg_types = {"this": True, "expression": True}


# Table-valued / UDTF


class GenerateSeries(Expression, Func):
    arg_types = {"start": True, "end": True, "step": False, "is_end_exclusive": False}


class ExplodingGenerateSeries(GenerateSeries):
    pass


class Generator(Expression, Func, UDTF):
    arg_types = {"rowcount": False, "timelimit": False}


class Explode(Expression, Func, UDTF):
    arg_types = {"this": True, "expressions": False}
    is_var_len_args = True


class Inline(Expression, Func):
    pass


@trait
class ExplodeOuter(Expr):
    pass


class _ExplodeOuter(Explode, ExplodeOuter):
    _sql_names = ["EXPLODE_OUTER"]


class Posexplode(Explode):
    pass


class PosexplodeOuter(Posexplode, ExplodeOuter):
    pass


class PositionalColumn(Expression):
    pass


class Unnest(Expression, Func, UDTF):
    arg_types = {
        "expressions": True,
        "alias": False,
        "offset": False,
        "explode_array": False,
    }

    @property
    def selects(self) -> list[Expr]:
        columns = super().selects
        offset = self.args.get("offset")
        if offset:
            columns = columns + [to_identifier("offset") if offset is True else offset]
        return columns


# Map


class Map(Expression, Func):
    arg_types = {"keys": False, "values": False}

    @property
    def keys(self) -> list[Expr]:
        keys = self.args.get("keys")
        return keys.expressions if keys else []

    @property
    def values(self) -> list[Expr]:
        values = self.args.get("values")
        return values.expressions if values else []


class MapCat(Expression, Func):
    arg_types = {"this": True, "expression": True}


class MapContainsKey(Expression, Func):
    arg_types = {"this": True, "key": True}


class MapDelete(Expression, Func):
    arg_types = {"this": True, "expressions": True}
    is_var_len_args = True


class MapFromEntries(Expression, Func):
    pass


class MapInsert(Expression, Func):
    arg_types = {"this": True, "key": False, "value": True, "update_flag": False}


class MapKeys(Expression, Func):
    pass


class MapPick(Expression, Func):
    arg_types = {"this": True, "expressions": True}
    is_var_len_args = True


class MapSize(Expression, Func):
    pass


class StarMap(Expression, Func):
    pass


class ToMap(Expression, Func):
    pass


class VarMap(Expression, Func):
    arg_types = {"keys": True, "values": True}
    is_var_len_args = True
    var_len_arg_key = "values"

    @property
    def keys(self) -> list[Expr]:
        return self.args["keys"].expressions

    @property
    def values(self) -> list[Expr]:
        return self.args["values"].expressions


# Struct


class Struct(Expression, Func):
    arg_types = {"expressions": False}
    is_var_len_args = True


class StructExtract(Expression, Func):
    arg_types = {"this": True, "expression": True}


# Geospatial


class StDistance(Expression, Func):
    arg_types = {"this": True, "expression": True, "use_spheroid": False}


class StPoint(Expression, Func):
    arg_types = {"this": True, "expression": True, "z": False, "m": False}
    _sql_names = ["ST_POINT", "ST_MAKEPOINT"]



from __future__ import annotations

import typing as t

from sqlglot import exp

# One recorded argument mutation: (node, arg_key, value before the mutation).
JournalEntry = tuple[exp.Expr, str, t.Any]
Journal = list[JournalEntry]


def record(journal: Journal, node: exp.Expr, arg_key: str) -> None:
    """Records the current value of `node.args[arg_key]` so `revert` can restore it.

    Must be called before the argument is mutated. List values are copied shallowly,
    so the caller may freely mutate or replace the list afterwards.
    """
    value = node.args.get(arg_key)
    journal.append((node, arg_key, list(value) if type(value) is list else value))


def revert(journal: Journal) -> None:
    """Restores every recorded argument, newest first, and empties the journal.

    `set` reattaches the restored expressions (their `parent`, `arg_key` and `index`),
    so rules that only detach or reorder nodes leave the tree exactly as it was before
    they ran. Rules that replace nodes with new ones can't be reverted this way.
    """
    for node, arg_key, value in reversed(journal):
        node.set(arg_key, value)

    journal.clear()



# ruff: noqa: F401
"""Lazy re-exports from optimizer submodules.

Eager re-exports here trip a circular import under sqlglot[c]: importing
sqlglot loads compiled `expressions.builders`, which eagerly wires up its
links to compiled optimizer modules, which causes Python to run this
file. The eager `from sqlglot.optimizer.optimizer import ...` then asks
for `sqlglot.Schema`, but `sqlglot/__init__.py` hasn't bound it yet.

PEP 562 `__getattr__` defers the import to first attribute access, by
which point sqlglot is fully loaded. Tracked upstream in python/mypy#21299.
"""

from __future__ import annotations

import threading
import typing as t

# Serialise concurrent first-access lookups; importlib's per-module lock is
# released before our caching write to globals(), so two threads racing on
# the same name could otherwise both run the import + getattr.
_import_lock = threading.RLock()

# Only for type checkers and IDEs; runtime resolution goes through __getattr__.
if t.TYPE_CHECKING:
    from sqlglot.optimizer.optimizer import RULES as RULES, optimize as optimize
    from sqlglot.optimizer.scope import (
        Scope as Scope,
        build_scope as build_scope,
        find_all_in_scope as find_all_in_scope,
        find_in_scope as find_in_scope,
        traverse_scope as traverse_scope,
        walk_in_scope as walk_in_scope,
    )

# Explicit, because some names collide between optimizer.py and submodules
# (e.g. `qualify` is both a function and a submodule).
_LAZY_ATTRS: dict[str, tuple[str, str]] = {
    "RULES": ("sqlglot.optimizer.optimizer", "RULES"),
    "optimize": ("sqlglot.optimizer.optimizer", "optimize"),
    "Scope": ("sqlglot.optimizer.scope", "Scope"),
    "build_scope": ("sqlglot.optimizer.scope", "build_scope"),
    "find_all_in_scope": ("sqlglot.optimizer.scope", "find_all_in_scope"),
    "find_in_scope": ("sqlglot.optimizer.scope", "find_in_scope"),
    "traverse_scope": ("sqlglot.optimizer.scope", "traverse_scope"),
    "walk_in_scope": ("sqlglot.optimizer.scope", "walk_in_scope"),
}


def __getattr__(name: str) -> t.Any:
    import importlib

    with _import_lock:
        target = _LAZY_ATTRS.get(name)
        if target is not None:
            module_name, attr = target
            value = getattr(importlib.import_module(module_name), attr)
        else:
            # Submodule fallback so `from sqlglot.optimizer import qualify` works.
            try:
                value = importlib.import_module(f"{__name__}.{name}")
            except ModuleNotFoundError:
                raise AttributeError(f"module {__name__!r} has no attribute {name!r}") from None
        globals()[name] = value
        return value



from __future__ import annotations
from sqlglot import exp
from sqlglot.helper import name_sequence
from sqlglot.optimizer.scope import ScopeType, find_in_scope, traverse_scope
from sqlglot._typing import E


def unnest_subqueries(expression: E) -> E:
    """
    Rewrite sqlglot AST to convert some predicates with subqueries into joins.

    Convert scalar subqueries into cross joins.
    Convert correlated or vectorized subqueries into a group by so it is not a many to many left join.

    Example:
        >>> import sqlglot
        >>> expression = sqlglot.parse_one("SELECT * FROM x AS x WHERE (SELECT y.a AS a FROM y AS y WHERE x.a = y.a) = 1 ")
        >>> unnest_subqueries(expression).sql()
        'SELECT * FROM x AS x LEFT JOIN (SELECT y.a AS a FROM y AS y WHERE TRUE GROUP BY y.a) AS _u_0 ON x.a = _u_0.a WHERE _u_0.a = 1'

    Args:
        expression (sqlglot.Expr): expression to unnest
    Returns:
        sqlglot.Expr: unnested expression
    """
    next_alias_name = name_sequence("_u_")

    for scope in traverse_scope(expression):
        select = scope.expression
        parent = select.parent_select
        if not parent:
            continue
        if scope.external_columns:
            decorrelate(select, parent, scope.external_columns, next_alias_name)
        elif scope.scope_type == ScopeType.SUBQUERY:
            unnest(select, parent, next_alias_name)

    return expression


def unnest(select, parent_select, next_alias_name):
    if len(select.selects) > 1:
        return

    predicate = select.find_ancestor(exp.Condition)
    if (
        not predicate
        # Do not unnest subqueries inside table-valued functions such as
        # FROM GENERATE_SERIES(...), FROM UNNEST(...) etc in order to preserve join order
        or (
            isinstance(predicate, exp.Func)
            and isinstance(predicate.parent, (exp.Table, exp.From, exp.Join))
        )
        or parent_select is not predicate.parent_select
        or not parent_select.args.get("from_")
        # NOT IN has three-valued semantics that the LEFT-JOIN-anti rewrite doesn't preserve:
        # a NULL in the subquery makes NOT IN evaluate to NULL for every outer row.
        or (isinstance(predicate, exp.In) and isinstance(predicate.parent, exp.Not))
    ):
        return

    if isinstance(select, exp.SetOperation):
        inner_alias = next_alias_name()
        select = exp.select(
            *(
                exp.alias_(exp.column(s.alias_or_name, inner_alias), s.alias_or_name)
                for s in select.selects
            )
        ).from_(select.subquery(inner_alias))

    alias = next_alias_name()
    clause = predicate.find_ancestor(exp.Having, exp.Where, exp.Join)

    # This subquery returns a scalar and can just be converted to a cross join
    if not isinstance(predicate, (exp.In, exp.Any)):
        column = exp.column(select.selects[0].alias_or_name, alias)

        clause_parent_select = clause.parent_select if clause else None

        if (isinstance(clause, exp.Having) and clause_parent_select is parent_select) or (
            (not clause or clause_parent_select is not parent_select)
            and (
                parent_select.args.get("group")
                or any(find_in_scope(select, exp.AggFunc) for select in parent_select.selects)
            )
        ):
            column = exp.Max(this=column)
        elif not isinstance(select.parent, exp.Subquery):
            return

        join_type = "CROSS"
        on_clause = None
        if isinstance(predicate, exp.Exists):
            # If a subquery returns no rows, cross-joining against it incorrectly eliminates all rows
            # from the parent query. Therefore, we use a LEFT JOIN that always matches (ON TRUE), then
            # check for non-NULL column values to determine whether the subquery contained rows.
            column = column.is_(exp.null()).not_()
            join_type = "LEFT"
            on_clause = exp.true()

        _replace(select.parent, column)
        parent_select.join(select, on=on_clause, join_type=join_type, join_alias=alias, copy=False)

        return

    if find_in_scope(select, exp.Limit, exp.Offset):
        return

    if isinstance(predicate, exp.Any):
        predicate = predicate.find_ancestor(exp.EQ)

        if not predicate or parent_select is not predicate.parent_select:
            return

    column = _other_operand(predicate)
    value = select.selects[0]

    join_key = exp.column(value.alias, alias)
    join_key_not_null = join_key.is_(exp.null()).not_()

    if isinstance(clause, exp.Join):
        _replace(predicate, exp.true())
        parent_select.where(join_key_not_null, copy=False)
    else:
        _replace(predicate, join_key_not_null)

    group = select.args.get("group")

    if group:
        if {value.this} != set(group.expressions):
            select = (
                exp.select(exp.alias_(exp.column(value.alias, "_q"), value.alias))
                .from_(select.subquery("_q", copy=False), copy=False)
                .group_by(exp.column(value.alias, "_q"), copy=False)
            )
    elif not find_in_scope(value.this, exp.AggFunc):
        select = select.group_by(value.this, copy=False)

    parent_select.join(
        select,
        on=column.eq(join_key),
        join_type="LEFT",
        join_alias=alias,
        copy=False,
    )


def decorrelate(select, parent_select, external_columns, next_alias_name):
    where = select.args.get("where")

    if not where or where.find(exp.Or) or select.find(exp.Limit, exp.Offset):
        return

    table_alias = next_alias_name()
    keys = []

    # for all external columns in the where statement, find the relevant predicate
    # keys to convert it into a join
    for column in external_columns:
        if column.find_ancestor(exp.Where) is not where:
            return

        # The predicate is replaced with TRUE below, which is only sound if its result flows
        # into the WHERE through conjunctions; wrappers like NOT would invert that TRUE.
        predicate = column.find_ancestor(exp.Predicate)
        ancestor = predicate.parent if predicate else None
        while isinstance(ancestor, (exp.And, exp.Paren)):
            ancestor = ancestor.parent

        if ancestor is not where:
            return

        if isinstance(predicate, exp.Binary):
            key = (
                predicate.right
                if any(node is column for node in predicate.left.walk())
                else predicate.left
            )
        else:
            return

        keys.append((key, column, predicate))

    if not any(isinstance(predicate, exp.EQ) for *_, predicate in keys):
        return

    is_subquery_projection = any(
        node is select.parent
        for node in map(lambda s: s.unalias(), parent_select.selects)
        if isinstance(node, exp.Subquery)
    )

    value = select.selects[0]
    key_aliases = {}
    group_by = []

    for key, _, predicate in keys:
        # if we filter on the value of the subquery, it needs to be unique
        if key == value.this:
            key_aliases[key] = value.alias
            group_by.append(key)
        else:
            if key not in key_aliases:
                key_aliases[key] = next_alias_name()
            # all predicates that are equalities must also be in the unique
            # so that we don't do a many to many join
            if isinstance(predicate, exp.EQ) and key not in group_by:
                group_by.append(key)

    parent_predicate = select.find_ancestor(exp.Predicate)

    # When the subquery is embedded inside a function (e.g. COALESCE, TRIM) in the SELECT list,
    # the ancestor chain contains no Predicate node AND the subquery is not a direct projection.
    if parent_predicate is None and not is_subquery_projection:
        return

    # if the value of the subquery is not an agg or a key, we need to collect it into an array
    # so that it can be grouped. For subquery projections, we use a MAX aggregation instead.
    agg_func = exp.Max if is_subquery_projection else exp.ArrayAgg
    if (
        not isinstance(value, exp.Subquery)
        and not find_in_scope(value, exp.AggFunc)
        and value.this not in group_by
    ):
        select.select(
            exp.alias_(agg_func(this=value.this), value.alias, quoted=False),
            append=False,
            copy=False,
        )

    # exists queries should not have any selects as it only checks if there are any rows
    # all selects will be added by the optimizer and only used for join keys
    if isinstance(parent_predicate, exp.Exists):
        select.set("expressions", [])

    for key, alias in key_aliases.items():
        if key in group_by:
            # add all keys to the projections of the subquery
            # so that we can use it as a join key
            if isinstance(parent_predicate, exp.Exists) or key != value.this:
                select.select(f"{key} AS {alias}", copy=False)
        else:
            select.select(exp.alias_(agg_func(this=key.copy()), alias, quoted=False), copy=False)

    alias = exp.column(value.alias, table_alias)
    other = _other_operand(parent_predicate)
    op_type = type(parent_predicate.parent) if parent_predicate else None

    if isinstance(parent_predicate, exp.Exists):
        alias = exp.column(list(key_aliases.values())[0], table_alias)
        parent_predicate = _replace(parent_predicate, f"NOT {alias} IS NULL")
    elif isinstance(parent_predicate, exp.All):
        assert issubclass(op_type, exp.Binary)
        predicate = op_type(this=other, expression=exp.column("_x"))
        parent_predicate = _replace(
            parent_predicate.parent, f"ARRAY_ALL({alias}, _x -> {predicate})"
        )
    elif isinstance(parent_predicate, exp.Any):
        assert issubclass(op_type, exp.Binary)
        if value.this in group_by:
            predicate = op_type(this=other, expression=alias)
            parent_predicate = _replace(parent_predicate.parent, predicate)
        else:
            predicate = op_type(this=other, expression=exp.column("_x"))
            parent_predicate = _replace(parent_predicate, f"ARRAY_ANY({alias}, _x -> {predicate})")
    elif isinstance(parent_predicate, exp.In):
        if value.this in group_by:
            parent_predicate = _replace(parent_predicate, f"{other} = {alias}")
        else:
            parent_predicate = _replace(
                parent_predicate,
                f"ARRAY_ANY({alias}, _x -> _x = {parent_predicate.this})",
            )
    else:
        if is_subquery_projection and select.parent.alias:
            alias = exp.alias_(alias, select.parent.alias)

        # COUNT always returns 0 on empty datasets, so we need take that into consideration here
        # by transforming all counts into 0 and using that as the coalesced value
        if find_in_scope(value, exp.Count):

            def remove_aggs(node):
                if isinstance(node, exp.Count):
                    return exp.Literal.number(0)
                elif isinstance(node, exp.AggFunc):
                    return exp.null()
                return node

            alias = exp.Coalesce(this=alias, expressions=[value.this.transform(remove_aggs)])

        select.parent.replace(alias)

    for key, column, predicate in keys:
        predicate.replace(exp.true())
        nested = exp.column(key_aliases[key], table_alias)

        if is_subquery_projection:
            key.replace(nested)
            if not isinstance(predicate, exp.EQ):
                parent_select.where(predicate, copy=False)
            continue

        if key in group_by:
            key.replace(nested)
        else:
            # Built as AST rather than a SQL string, because dialect-specific operators such as
            # Postgres' `@>` can't be round-tripped through the default dialect's parser.
            key.replace(exp.to_identifier("_x"))
            right = exp.ArrayAny(
                this=nested,
                expression=exp.Lambda(this=predicate.copy(), expressions=[exp.to_identifier("_x")]),
            )
            parent_predicate = _replace(
                parent_predicate,
                exp.paren(exp.and_(parent_predicate.copy(), right, copy=False)),
            )

    parent_select.join(
        select.group_by(*group_by, copy=False),
        on=[predicate for *_, predicate in keys if isinstance(predicate, exp.EQ)],
        join_type="LEFT",
        join_alias=table_alias,
        copy=False,
    )


def _replace(expression: exp.Expr, condition: exp.ExpOrStr) -> exp.Expr:
    return expression.replace(exp.condition(condition))


def _other_operand(expression: object) -> exp.Expr | None:
    if isinstance(expression, exp.In):
        return expression.this

    if isinstance(expression, (exp.Any, exp.All)):
        return _other_operand(expression.parent)

    if isinstance(expression, exp.Binary):
        return (
            expression.right
            if isinstance(expression.left, (exp.Subquery, exp.Any, exp.Exists, exp.All))
            else expression.left
        )

    return None



"""sqlglot expressions datatypes."""

from __future__ import annotations

import typing as t
from enum import auto

from sqlglot.helper import AutoName
from sqlglot.errors import ErrorLevel, ParseError
from sqlglot.expressions.core import (
    Expr,
    Expression,
    _TimeUnit,
    Identifier,
    Dot,
    maybe_copy,
)
from builtins import type as Type

if t.TYPE_CHECKING:
    from sqlglot._typing import DataTypeArgs
    from sqlglot.dialects.dialect import DialectType
    from typing_extensions import Self, Unpack


class DataTypeParam(Expression):
    arg_types = {"this": True, "expression": False}

    @property
    def this(self) -> Expr:
        return self.args["this"]

    @property
    def name(self) -> str:
        return self.this.name


class DType(AutoName):
    ARRAY = auto()
    AGGREGATEFUNCTION = auto()
    SIMPLEAGGREGATEFUNCTION = auto()
    BIGDECIMAL = auto()
    BIGINT = auto()
    BIGNUM = auto()
    BIGSERIAL = auto()
    BINARY = auto()
    BIT = auto()
    BLOB = auto()
    BOOLEAN = auto()
    BPCHAR = auto()
    CHAR = auto()
    CHARACTER_SET = auto()
    DATE = auto()
    DATE32 = auto()
    DATEMULTIRANGE = auto()
    DATERANGE = auto()
    DATETIME = auto()
    DATETIME2 = auto()
    DATETIME64 = auto()
    DECIMAL = auto()
    DECIMAL32 = auto()
    DECIMAL64 = auto()
    DECIMAL128 = auto()
    DECIMAL256 = auto()
    DECFLOAT = auto()
    DOUBLE = auto()
    DYNAMIC = auto()
    ENUM = auto()
    ENUM8 = auto()
    ENUM16 = auto()
    FILE = auto()
    FIXEDSTRING = auto()
    FLOAT = auto()
    GEOGRAPHY = auto()
    GEOGRAPHYPOINT = auto()
    GEOMETRY = auto()
    POINT = auto()
    RING = auto()
    LINESTRING = auto()
    MULTILINESTRING = auto()
    POLYGON = auto()
    MULTIPOLYGON = auto()
    HLLSKETCH = auto()
    HSTORE = auto()
    IMAGE = auto()
    INET = auto()
    INT = auto()
    INT128 = auto()
    INT256 = auto()
    INT4MULTIRANGE = auto()
    INT4RANGE = auto()
    INT8MULTIRANGE = auto()
    INT8RANGE = auto()
    INTERVAL = auto()
    IPADDRESS = auto()
    IPPREFIX = auto()
    IPV4 = auto()
    IPV6 = auto()
    JSON = auto()
    JSONB = auto()
    LIST = auto()
    LONGBLOB = auto()
    LONGTEXT = auto()
    LOWCARDINALITY = auto()
    MAP = auto()
    MEDIUMBLOB = auto()
    MEDIUMINT = auto()
    MEDIUMTEXT = auto()
    MONEY = auto()
    NAME = auto()
    NCHAR = auto()
    NESTED = auto()
    NOTHING = auto()
    NULL = auto()
    NUMMULTIRANGE = auto()
    NUMRANGE = auto()
    NVARCHAR = auto()
    OBJECT = auto()
    RANGE = auto()
    ROWVERSION = auto()
    SERIAL = auto()
    SET = auto()
    SMALLDATETIME = auto()
    SMALLINT = auto()
    SMALLMONEY = auto()
    SMALLSERIAL = auto()
    STRUCT = auto()
    SUPER = auto()
    TEXT = auto()
    TINYBLOB = auto()
    TINYTEXT = auto()
    TIME = auto()
    TIMETZ = auto()
    TIME_NS = auto()
    TIMESTAMP = auto()
    TIMESTAMPNTZ = auto()
    TIMESTAMPLTZ = auto()
    TIMESTAMPTZ = auto()
    TIMESTAMP_S = auto()
    TIMESTAMP_MS = auto()
    TIMESTAMP_NS = auto()
    TINYINT = auto()
    TSMULTIRANGE = auto()
    TSRANGE = auto()
    TSTZMULTIRANGE = auto()
    TSTZRANGE = auto()
    UBIGINT = auto()
    UINT = auto()
    UINT128 = auto()
    UINT256 = auto()
    UMEDIUMINT = auto()
    UDECIMAL = auto()
    UDOUBLE = auto()
    UNION = auto()
    UNKNOWN = auto()  # Sentinel value, useful for type annotation
    USERDEFINED = "USER-DEFINED"
    USMALLINT = auto()
    UTINYINT = auto()
    UUID = auto()
    VARBINARY = auto()
    VARCHAR = auto()
    VARIANT = auto()
    VECTOR = auto()
    XML = auto()
    YEAR = auto()
    TDIGEST = auto()

    def into_expr(self, **kwargs: object) -> DataType:
        """Converts this `DType` into a `DataType` instance.

        Args:
            **kwargs (object): additional arguments to pass in the constructor of DataType.
        Returns:
            DataType: the resulting `DataType` instance.
        """
        return DataType(this=self).set_kwargs(kwargs)


class DataType(Expression):
    arg_types = {
        "this": True,
        "expressions": False,
        "nested": False,
        "values": False,
        "kind": False,
        "nullable": False,
        "collate": False,
    }

    is_data_type: t.ClassVar[bool] = True

    Type: t.ClassVar[Type[DType]] = DType

    STRUCT_TYPES: t.ClassVar[set[DType]] = {
        DType.FILE,
        DType.NESTED,
        DType.OBJECT,
        DType.STRUCT,
        DType.UNION,
    }

    ARRAY_TYPES: t.ClassVar[set[DType]] = {
        DType.ARRAY,
        DType.LIST,
    }

    NESTED_TYPES: t.ClassVar[set[DType]] = {
        DType.FILE,
        DType.NESTED,
        DType.OBJECT,
        DType.STRUCT,
        DType.UNION,
        DType.ARRAY,
        DType.LIST,
        DType.MAP,
    }

    TEXT_TYPES: t.ClassVar[set[DType]] = {
        DType.CHAR,
        DType.NCHAR,
        DType.NVARCHAR,
        DType.TEXT,
        DType.TINYTEXT,
        DType.MEDIUMTEXT,
        DType.LONGTEXT,
        DType.VARCHAR,
        DType.NAME,
    }

    BINARY_TYPES: t.ClassVar[set[DType]] = {
        DType.BINARY,
        DType.VARBINARY,
        DType.TINYBLOB,
        DType.BLOB,
        DType.MEDIUMBLOB,
        DType.LONGBLOB,
    }

    SIGNED_INTEGER_TYPES: t.ClassVar[set[DType]] = {
        DType.BIGINT,
        DType.INT,
        DType.INT128,
        DType.INT256,
        DType.MEDIUMINT,
        DType.SMALLINT,
        DType.TINYINT,
    }

    UNSIGNED_INTEGER_TYPES: t.ClassVar[set[DType]] = {
        DType.UBIGINT,
        DType.UINT,
        DType.UINT128,
        DType.UINT256,
        DType.UMEDIUMINT,
        DType.USMALLINT,
        DType.UTINYINT,
    }

    INTEGER_TYPES: t.ClassVar[set[DType]] = {
        DType.BIGINT,
        DType.INT,
        DType.INT128,
        DType.INT256,
        DType.MEDIUMINT,
        DType.SMALLINT,
        DType.TINYINT,
        DType.UBIGINT,
        DType.UINT,
        DType.UINT128,
        DType.UINT256,
        DType.UMEDIUMINT,
        DType.USMALLINT,
        DType.UTINYINT,
        DType.BIT,
    }

    FLOAT_TYPES: t.ClassVar[set[DType]] = {
        DType.DOUBLE,
        DType.FLOAT,
    }

    REAL_TYPES: t.ClassVar[set[DType]] = {
        DType.DOUBLE,
        DType.FLOAT,
        DType.BIGDECIMAL,
        DType.DECIMAL,
        DType.DECIMAL32,
        DType.DECIMAL64,
        DType.DECIMAL128,
        DType.DECIMAL256,
        DType.DECFLOAT,
        DType.MONEY,
        DType.SMALLMONEY,
        DType.UDECIMAL,
        DType.UDOUBLE,
    }

    NUMERIC_TYPES: t.ClassVar[set[DType]] = {
        DType.BIGINT,
        DType.INT,
        DType.INT128,
        DType.INT256,
        DType.MEDIUMINT,
        DType.SMALLINT,
        DType.TINYINT,
        DType.UBIGINT,
        DType.UINT,
        DType.UINT128,
        DType.UINT256,
        DType.UMEDIUMINT,
        DType.USMALLINT,
        DType.UTINYINT,
        DType.BIT,
        DType.DOUBLE,
        DType.FLOAT,
        DType.BIGDECIMAL,
        DType.DECIMAL,
        DType.DECIMAL32,
        DType.DECIMAL64,
        DType.DECIMAL128,
        DType.DECIMAL256,
        DType.DECFLOAT,
        DType.MONEY,
        DType.SMALLMONEY,
        DType.UDECIMAL,
        DType.UDOUBLE,
    }

    TEMPORAL_TYPES: t.ClassVar[set[DType]] = {
        DType.DATE,
        DType.DATE32,
        DType.DATETIME,
        DType.DATETIME2,
        DType.DATETIME64,
        DType.SMALLDATETIME,
        DType.TIME,
        DType.TIMESTAMP,
        DType.TIMESTAMPNTZ,
        DType.TIMESTAMPLTZ,
        DType.TIMESTAMPTZ,
        DType.TIMESTAMP_MS,
        DType.TIMESTAMP_NS,
        DType.TIMESTAMP_S,
        DType.TIMETZ,
    }

    @classmethod
    def build(
        cls,
        dtype: DATA_TYPE,
        dialect: DialectType = None,
        udt: bool = False,
        copy: bool = True,
        **kwargs: Unpack[DataTypeArgs],
    ) -> Self:
        """
        Constructs a DataType object.

        Args:
            dtype: the data type of interest.
            dialect: the dialect to use for parsing `dtype`, in case it's a string.
            udt: when set to True, `dtype` will be used as-is if it can't be parsed into a
                DataType, thus creating a user-defined type.
            copy: whether to copy the data type.
            kwargs: additional arguments to pass in the constructor of DataType.

        Returns:
            The constructed DataType object.
        """
        if isinstance(dtype, str):
            return cls.from_str(dtype, dialect, udt, **kwargs)
        elif isinstance(dtype, DType):
            data_type_exp = cls(this=dtype)
            if kwargs:
                for k, v in kwargs.items():
                    data_type_exp.set(k, v)
            return data_type_exp
        elif isinstance(dtype, (Identifier, Dot)) and udt:
            return cls(this=DType.USERDEFINED, kind=dtype, **kwargs)
        elif isinstance(dtype, cls):
            return maybe_copy(dtype, copy)
        else:
            raise ValueError(f"Invalid data type: {type(dtype)}. Expected str or DType")

    @classmethod
    def from_str(
        cls,
        dtype: str,
        dialect: DialectType = None,
        udt: bool = False,
        **kwargs: Unpack[DataTypeArgs],
    ) -> Self:
        """
        Constructs a `DataType` object from a `str` representation.

        Args:
            dtype: the data type of interest.
            dialect: the dialect to use for parsing `dtype`.
            udt: when set to True, `dtype` will be used as-is if it can't be parsed into a
                `DataType`, thus creating a user-defined type.
            kwargs: additional arguments to pass in the constructor of `DataType`.

        Returns:
            The constructed `DataType` object.
        """
        from sqlglot import parse_one

        if dtype.upper() == "UNKNOWN":
            return cls(this=DType.UNKNOWN, **kwargs)
        try:
            return parse_one(
                dtype, read=dialect, into=cls, error_level=ErrorLevel.IGNORE
            ).set_kwargs(kwargs)
        except ParseError:
            if udt:
                return cls(this=DType.USERDEFINED, kind=dtype, **kwargs)
            raise

    def is_type(self, *dtypes: DATA_TYPE, check_nullable: bool = False) -> bool:
        """
        Checks whether this DataType matches one of the provided data types. Nested types or precision
        will be compared using "structural equivalence" semantics, so e.g. array<int> != array<float>.

        Args:
            dtypes: the data types to compare this DataType to.
            check_nullable: whether to take the NULLABLE type constructor into account for the comparison.
                If false, it means that NULLABLE<INT> is equivalent to INT.

        Returns:
            True, if and only if there is a type in `dtypes` which is equal to this DataType.
        """
        self_is_nullable: bool | None = self.args.get("nullable")
        for dtype in dtypes:
            other_type = DataType.build(dtype, copy=False, udt=True)
            other_is_nullable: bool | None = other_type.args.get("nullable")
            if (
                other_type.expressions
                or (check_nullable and (self_is_nullable or other_is_nullable))
                or self.this == DType.USERDEFINED
                or other_type.this == DType.USERDEFINED
            ):
                matches = self == other_type
            else:
                matches = self.this == other_type.this

            if matches:
                return True
        return False


class PseudoType(DataType):
    arg_types = {"this": True}


class ObjectIdentifier(DataType):
    arg_types = {"this": True}


class IntervalSpan(DataType):
    arg_types = {"this": True, "expression": True}


class Interval(_TimeUnit):
    arg_types = {"this": False, "unit": False}


DATA_TYPE = t.Union[str, Identifier, Dot, DataType, DType]



from __future__ import annotations

from sqlglot import exp
from sqlglot.typing import EXPRESSION_METADATA

EXPRESSION_METADATA = {
    **EXPRESSION_METADATA,
    **{
        expr_type: {"returns": exp.DType.BIGINT}
        for expr_type in {
            exp.BitwiseAnd,
            exp.BitwiseNot,
            exp.BitwiseOr,
            exp.BitwiseXor,
            exp.Length,
            exp.Levenshtein,
            exp.StrPosition,
            exp.WidthBucket,
        }
    },
    **{
        expr_type: {"annotator": lambda self, e: self._annotate_by_args(e, "this")}
        for expr_type in {
            exp.Ceil,
            exp.Floor,
            exp.Round,
            exp.Sign,
        }
    },
    exp.Mod: {"annotator": lambda self, e: self._annotate_by_args(e, "this", "expression")},
    exp.Rand: {
        "annotator": lambda self, e: (
            self._annotate_by_args(e, "this") if e.this else self._set_type(e, exp.DType.DOUBLE)
        )
    },
    exp.MD5Digest: {"returns": exp.DType.VARBINARY},
}
